from __future__ import annotations

from datetime import timezone
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

import arrow
from btrfs2s3._internal.assessor import Assessment
from btrfs2s3._internal.assessor import BackupAssessment
from btrfs2s3._internal.assessor import SnapshotAssessment
from btrfs2s3._internal.assessor import SourceAssessment
from btrfs2s3._internal.commands.update import print_assessment
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.thunk import Thunk
from btrfs2s3._internal.util import backup_of_snapshot
from btrfs2s3._internal.util import mksubvol
from btrfs2s3._internal.util import SubvolumeFlags
import pytest

if TYPE_CHECKING:
    from rich.console import Console


print_assessment_under_test = partial(
    print_assessment,
    tzinfo=timezone.utc,
    snapshot_dir=Path("/snapshots"),
    bucket="my-bucket",
)


@pytest.fixture()
def golden_asmt_with_no_changes() -> Assessment:
    asmt = Assessment()
    source1 = mksubvol(
        id=12345,
        uuid=b"$\x96\xea\xb9\x1f\xbcD\x7f\xba\x98\xf1\xc5b\xaf\xc6A",
        ctransid=123,
        ctime=arrow.get("2006-01-02").timestamp(),
    )
    source1_asmt = SourceAssessment(path=Path("/path/to/source1"), info=source1)
    asmt.sources[source1.uuid] = source1_asmt
    snap1_1 = mksubvol(
        id=source1.id + 1,
        uuid=b"4k\xd4D\xb1\x87O\xdb\x88\xdd=\xa6\xfd\xef\x84B",
        parent_uuid=source1.uuid,
        flags=SubvolumeFlags.ReadOnly,
        ctransid=source1.ctransid - 3,
        ctime=source1.ctime - 86400,
    )
    source1_asmt.snapshots[snap1_1.uuid] = SnapshotAssessment(
        initial_path=Path("/snapshots/snap1_1"),
        info=snap1_1,
        target_path=Thunk(Path("/snapshots/snap1_1")),
        real_info=Thunk(snap1_1),
        keep_meta=KeepMeta(
            reasons=Reasons.Preserved,
            time_spans={(arrow.get("2006").timestamp(), arrow.get("2007").timestamp())},
        ),
    )
    snap1_2 = mksubvol(
        id=snap1_1.id + 1,
        uuid=b"\x11[\xab\xa0\xec\xedA\xa6\xbbm3D\x9c\x92\xbf=",
        parent_uuid=source1.uuid,
        flags=SubvolumeFlags.ReadOnly,
        ctransid=source1.ctransid - 1,
        ctime=source1.ctime,
    )
    source1_asmt.snapshots[snap1_2.uuid] = SnapshotAssessment(
        initial_path=Path("/snapshots/snap1_2"),
        info=snap1_2,
        target_path=Thunk(Path("/snapshots/snap1_2")),
        real_info=Thunk(snap1_2),
        keep_meta=source1_asmt.snapshots[snap1_1.uuid].keep_meta,
    )
    backup1_1 = backup_of_snapshot(snap1_1, send_parent=None)
    source1_asmt.backups[backup1_1.uuid] = BackupAssessment(
        backup=Thunk(backup1_1),
        key=Thunk("backup1_1"),
        keep_meta=source1_asmt.snapshots[snap1_1.uuid].keep_meta,
    )
    backup1_2 = backup_of_snapshot(snap1_2, send_parent=snap1_1)
    source1_asmt.backups[backup1_2.uuid] = BackupAssessment(
        backup=Thunk(backup1_2),
        key=Thunk("backup1_2"),
        keep_meta=source1_asmt.snapshots[snap1_2.uuid].keep_meta,
    )
    return asmt


def test_print_assessment(
    golden_asmt_with_no_changes: Assessment, goldifyconsole: Console
) -> None:
    print_assessment_under_test(
        asmt=golden_asmt_with_no_changes, console=goldifyconsole
    )


def test_dont_keep(
    golden_asmt_with_no_changes: Assessment, goldifyconsole: Console
) -> None:
    # Pick an arbitrary, stable source
    _, source_asmt = sorted(golden_asmt_with_no_changes.sources.items())[0]
    # Pick an arbitrary, stable snapshot
    _, snapshot_asmt = sorted(source_asmt.snapshots.items())[0]
    # Mark the snapshot as not kept
    snapshot_asmt.keep_meta = KeepMeta()
    # Associated backup is not kept either
    backup_asmt = source_asmt.backups[snapshot_asmt.info.uuid]
    backup_asmt.keep_meta = KeepMeta()

    print_assessment_under_test(
        asmt=golden_asmt_with_no_changes, console=goldifyconsole
    )


def test_most_recent(
    golden_asmt_with_no_changes: Assessment, goldifyconsole: Console
) -> None:
    # Pick an arbitrary, stable source
    _, source_asmt = sorted(golden_asmt_with_no_changes.sources.items())[0]
    # Pick an arbitrary, stable snapshot
    _, snapshot_asmt = sorted(source_asmt.snapshots.items())[0]
    # Mark the snapshot as kept due to being the most recent
    snapshot_asmt.keep_meta = KeepMeta(reasons=Reasons.MostRecent)
    # Associated backup is kept for the same reason
    backup_asmt = source_asmt.backups[snapshot_asmt.info.uuid]
    backup_asmt.keep_meta = KeepMeta(reasons=Reasons.MostRecent)

    print_assessment_under_test(
        asmt=golden_asmt_with_no_changes, console=goldifyconsole
    )


def test_new_proposed(
    golden_asmt_with_no_changes: Assessment, goldifyconsole: Console
) -> None:
    # Pick an arbitrary, stable source
    _, source_asmt = sorted(golden_asmt_with_no_changes.sources.items())[0]
    # Create a new proposed snapshot
    proposed = mksubvol(
        id=999,
        uuid=b"\xda\x07m\x9ex\xfaLw\x89\xafYF>\xdc1\xf4",
        parent_uuid=source_asmt.info.uuid,
        flags=SubvolumeFlags.ReadOnly | SubvolumeFlags.Proposed,
        ctransid=source_asmt.info.ctransid,
        ctime=source_asmt.info.ctime,
    )
    source_asmt.snapshots[proposed.uuid] = SnapshotAssessment(
        initial_path=Path("/snapshots/proposed"),
        info=proposed,
        target_path=Thunk(lambda: Path("/snapshots/snap1_3")),  # pragma: no cover
        real_info=Thunk(lambda: proposed),  # pragma: no cover
        keep_meta=KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New),
    )
    # And associated backup
    backup = backup_of_snapshot(proposed, send_parent=None)
    source_asmt.backups[backup.uuid] = BackupAssessment(
        backup=Thunk(lambda: backup),  # pragma: no cover
        key=Thunk(lambda: "backup1-3"),  # pragma: no cover
        keep_meta=source_asmt.snapshots[proposed.uuid].keep_meta,
    )

    print_assessment_under_test(
        asmt=golden_asmt_with_no_changes, console=goldifyconsole
    )


def test_backup_without_snapshot(
    golden_asmt_with_no_changes: Assessment, goldifyconsole: Console
) -> None:
    # Pick an arbitrary, stable source
    _, source_asmt = sorted(golden_asmt_with_no_changes.sources.items())[0]
    # Pick an arbitrary, stable snapshot
    uuid, _ = sorted(source_asmt.snapshots.items())[0]
    source_asmt.snapshots.pop(uuid)

    print_assessment_under_test(
        asmt=golden_asmt_with_no_changes, console=goldifyconsole
    )


def test_send_ancestor(
    golden_asmt_with_no_changes: Assessment, goldifyconsole: Console
) -> None:
    # Pick an arbitrary, stable source
    _, source_asmt = sorted(golden_asmt_with_no_changes.sources.items())[0]
    # Pick an arbitrary, stable backup which has a send-parent
    child_uuid, child_backup_asmt = sorted(
        source_asmt.backups.items(),
        key=lambda i: (i[1].backup().send_parent_uuid is None, i[0]),
    )[0]
    # Get its parent snapshot and backup
    parent_uuid = child_backup_asmt.backup().send_parent_uuid
    assert parent_uuid is not None
    parent_snapshot_asmt = source_asmt.snapshots[parent_uuid]
    parent_backup_asmt = source_asmt.backups[parent_uuid]
    # Snapshot is not kept
    parent_snapshot_asmt.keep_meta = KeepMeta()
    # Backup must be kept because it's the send-parent of another backup
    parent_backup_asmt.keep_meta = KeepMeta(
        reasons=Reasons.SendAncestor, other_uuids={child_uuid}
    )

    print_assessment_under_test(
        asmt=golden_asmt_with_no_changes, console=goldifyconsole
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from btrfs2s3._internal.action import delete_backups

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client


def test_good_delete_one_key(s3: S3Client, bucket: str) -> None:
    key = "test-backup"
    s3.put_object(Bucket=bucket, Key=key, Body=b"dummy")

    delete_backups(s3, bucket, key)

    assert s3.list_objects_v2(Bucket=bucket).get("Contents", []) == []


def test_delete_key_that_does_not_exist(s3: S3Client, bucket: str) -> None:
    key = "test-backup"

    delete_backups(s3, bucket, key)

    assert s3.list_objects_v2(Bucket=bucket).get("Contents", []) == []


def test_some_keys_exist_and_some_do_not(s3: S3Client, bucket: str) -> None:
    key1 = "test-backup1"
    key2 = "test-backup2"
    s3.put_object(Bucket=bucket, Key=key1, Body=b"dummy")

    delete_backups(s3, bucket, key1, key2)

    assert s3.list_objects_v2(Bucket=bucket).get("Contents", []) == []


def test_large_numbers_of_keys(s3: S3Client, bucket: str) -> None:
    keys = [f"test-backup{i}" for i in range(1001)]
    for key in keys:
        s3.put_object(Bucket=bucket, Key=key, Body=b"dummy")

    delete_backups(s3, bucket, *keys)

    assert s3.list_objects_v2(Bucket=bucket).get("Contents", []) == []



"""The global rich text console."""

from __future__ import annotations

from rich.console import Console
from rich.style import Style
from rich.theme import Theme

STYLE_KEY = Style.parse("green")
STYLE_KEEP = Style.parse("green")
STYLE_NOT_KEEPING = Style.parse("bold bright_red")
STYLE_CTRANSID = Style.parse("cyan")

STYLE_CREATE = Style.parse("green")
STYLE_MODIFY = Style.parse("green")
STYLE_DELETE = Style.parse("red")

THEME = Theme(
    {
        "key": STYLE_KEY,
        "keep": STYLE_KEEP,
        "not_keeping": STYLE_NOT_KEEPING,
        "ctransid": STYLE_CTRANSID,
        "create": STYLE_CREATE,
        "modify": STYLE_MODIFY,
        "delete": STYLE_DELETE,
    }
)

CONSOLE = Console(theme=THEME)
"""The global rich text console."""



"""Functions for uploading a non-seekable stream to S3."""

from __future__ import annotations

import os
from tempfile import TemporaryFile
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import AnyStr
    from typing import IO
    from typing import Iterator

    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import CompletedPartTypeDef

_COPY_BUFFER_SIZE = 2**20


def _copy(*, input_file: IO[bytes], output_file: IO[bytes], count: int) -> int:
    # notes for fast copy:
    # - buffered IO classes have seek position out of sync with underlying file
    #   descriptor, mitigate this with SEEK_END
    # - sendfile and splice. which should be used first?
    # - use readinto and reuse buffers for slow mode
    written = 0
    eof = False
    while written < count and not eof:
        # This may result in one or multiple underlying reads
        buf = input_file.read(min(count - written, _COPY_BUFFER_SIZE))
        if not buf:
            eof = True
            break
        # https://docs.python.org/3/library/io.html#io.RawIOBase.write
        # indicates this loop is needed
        offset = 0
        while offset < len(buf):
            offset += output_file.write(buf[offset:])
        written += len(buf)
    return written


def _iter_parts_via_tempfile(stream: IO[bytes], part_size: int) -> Iterator[IO[bytes]]:
    while True:
        with TemporaryFile() as part_file:
            written = _copy(input_file=stream, output_file=part_file, count=part_size)
            if written > 0:
                part_file.seek(0, os.SEEK_SET)
                yield part_file
            if written < part_size:
                break


def _stream_len(stream: IO[AnyStr]) -> int:
    cur = stream.tell()
    end = stream.seek(0, os.SEEK_END)
    stream.seek(cur, os.SEEK_SET)
    return end


def upload_non_seekable_stream_via_tempfile(
    *, stream: IO[bytes], client: S3Client, bucket: str, key: str, part_size: int
) -> None:
    """Upload a non-seekable stream to S3.

    This will store the stream in parts to temporary files, of part_size bytes
    each. If less than one full part is consumed from the stream, it will
    upload the object with put_object. Otherwise, a multipart upload will be
    used.

    If any error is raised, this function will attempt to cancel the multipart
    upload with abort_multipart_upload().

    Args:
        stream: A stream to upload. The stream may be seekable, but this
            function is designed for the non-seekable case.
        client: The S3 client object.
        bucket: The name of the S3 bucket.
        key: The key of the S3 object in the bucket.
        part_size: The maximum size of a single part.
    """
    # If the first part is the maximum part size, assume there will be more parts. This
    # is suboptimal in the rare case that the stream is exactly one part length long.
    # The alternative is to attempt to read an extra byte from the stream after the
    # first part has been collected, and append it to the next part. The 1-byte reads
    # will frequently be unaligned and lead to cache thrashing. The optimal strategy
    # would be:
    # - Read the first full part
    # - Read 1 test byte
    # - Read the second part minus one byte
    # - Read the remaining parts as normal
    # This would be a lot of code complexity for a very rare gain.
    upload_id: str | None = None
    completed_parts: list[CompletedPartTypeDef] = []
    try:
        for part_index, part_file in enumerate(
            _iter_parts_via_tempfile(stream, part_size)
        ):
            if upload_id is None and _stream_len(part_file) == part_size:
                upload_id = client.create_multipart_upload(Bucket=bucket, Key=key)[
                    "UploadId"
                ]
            if upload_id is not None:
                part_number = part_index + 1
                up_response = client.upload_part(
                    Bucket=bucket,
                    Key=key,
                    PartNumber=part_number,
                    UploadId=upload_id,
                    Body=part_file,
                )
                completed_parts.append(
                    {"ETag": up_response["ETag"], "PartNumber": part_number}
                )
            else:
                client.put_object(Bucket=bucket, Key=key, Body=part_file)
        if upload_id is not None:
            client.complete_multipart_upload(
                Bucket=bucket,
                Key=key,
                UploadId=upload_id,
                MultipartUpload={"Parts": completed_parts},
            )
    except Exception:
        if upload_id is not None:
            client.abort_multipart_upload(Bucket=bucket, Key=key, UploadId=upload_id)
        raise



from pathlib import Path

from btrfs2s3._internal.action import create_snapshot
from btrfs2s3._internal.util import SubvolumeFlags
import btrfsutil


def test_call(btrfs_mountpoint: Path) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    source_info = btrfsutil.subvolume_info(source)
    path = btrfs_mountpoint / "snapshot"

    create_snapshot(source=source, path=path)

    assert btrfsutil.is_subvolume(path)
    path_info = btrfsutil.subvolume_info(path)
    assert path_info.parent_uuid == source_info.uuid
    assert path_info.flags & SubvolumeFlags.ReadOnly



from __future__ import annotations

from typing import Protocol
from typing import TYPE_CHECKING
from uuid import uuid4

import arrow
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.resolver import _Resolver
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepBackup
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.util import backup_of_snapshot
from btrfs2s3._internal.util import mksubvol
import pytest

if TYPE_CHECKING:
    from btrfsutil import SubvolumeInfo


@pytest.fixture()
def parent_uuid() -> bytes:
    return uuid4().bytes


class MkSnap(Protocol):
    def __call__(self, *, t: str | None = None, i: int = 0) -> SubvolumeInfo: ...


@pytest.fixture()
def mksnap(parent_uuid: bytes) -> MkSnap:
    def inner(*, t: str | None = None, i: int = 0) -> SubvolumeInfo:
        a = arrow.get() if t is None else arrow.get(t)
        return mksubvol(
            uuid=uuid4().bytes, parent_uuid=parent_uuid, ctime=a.timestamp(), ctransid=i
        )

    return inner


def test_simple_backup_of_snapshot(mksnap: MkSnap) -> None:
    snapshot = mksnap()
    resolver = _Resolver(snapshots=(snapshot,), backups=(), policy=Policy())

    with resolver._with_reasons(Reasons.Preserved):
        got = resolver._keep_backup_of_snapshot(snapshot)

    expected = backup_of_snapshot(snapshot, send_parent=None)
    assert got == expected

    assert resolver.get_result().keep_backups == {
        got.uuid: KeepBackup(got, KeepMeta(reasons=Reasons.Preserved, flags=Flags.New))
    }


def test_backup_done_twice(mksnap: MkSnap) -> None:
    snapshot = mksnap()
    resolver = _Resolver(snapshots=(snapshot,), backups=(), policy=Policy())

    with resolver._with_reasons(Reasons.Preserved):
        got1 = resolver._keep_backup_of_snapshot(snapshot)
    with resolver._with_reasons(Reasons.Preserved):
        got2 = resolver._keep_backup_of_snapshot(snapshot, flags=Flags.New)

    assert got1 == got2

    assert resolver.get_result().keep_backups == {
        got1.uuid: KeepBackup(
            got1, KeepMeta(reasons=Reasons.Preserved, flags=Flags.New)
        )
    }


def test_choose_correct_parents(mksnap: MkSnap) -> None:
    snapshot1 = mksnap(t="2006-01-01", i=1)
    snapshot2 = mksnap(t="2006-02-01", i=2)
    snapshot3 = mksnap(t="2006-02-01", i=3)
    snapshot4 = mksnap(t="2006-02-02", i=4)
    resolver = _Resolver(
        snapshots=(snapshot1, snapshot2, snapshot3, snapshot4),
        backups=(),
        policy=Policy.all(),
    )

    with resolver._with_reasons(Reasons.Preserved):
        backup1 = resolver._keep_backup_of_snapshot(snapshot1)
        backup2 = resolver._keep_backup_of_snapshot(snapshot2)
        backup3 = resolver._keep_backup_of_snapshot(snapshot3)
        backup4 = resolver._keep_backup_of_snapshot(snapshot4)

    expected1 = backup_of_snapshot(snapshot1, send_parent=None)
    assert backup1 == expected1
    expected2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    assert backup2 == expected2
    expected3 = backup_of_snapshot(snapshot3, send_parent=snapshot2)
    assert backup3 == expected3
    expected4 = backup_of_snapshot(snapshot4, send_parent=snapshot2)
    assert backup4 == expected4

    assert resolver.get_result().keep_backups == {
        backup1.uuid: KeepBackup(
            backup1, KeepMeta(reasons=Reasons.Preserved, flags=Flags.New)
        ),
        backup2.uuid: KeepBackup(
            backup2, KeepMeta(reasons=Reasons.Preserved, flags=Flags.New)
        ),
        backup3.uuid: KeepBackup(
            backup3, KeepMeta(reasons=Reasons.Preserved, flags=Flags.New)
        ),
        backup4.uuid: KeepBackup(
            backup4, KeepMeta(reasons=Reasons.Preserved, flags=Flags.New)
        ),
    }


def test_existing_backup(mksnap: MkSnap) -> None:
    snapshot1 = mksnap(t="2006-01-01", i=1)
    snapshot2 = mksnap(t="2006-02-01", i=2)
    # full backup, as normal
    backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    # full backup wouldn't normally be done here, but ensure we utilize this
    # existing backup if we need to backup the snapshot
    backup2 = backup_of_snapshot(snapshot2, send_parent=None)
    resolver = _Resolver(
        snapshots=(snapshot1, snapshot2), backups=(backup1, backup2), policy=Policy()
    )

    with resolver._with_reasons(Reasons.Preserved):
        got_backup1 = resolver._keep_backup_of_snapshot(snapshot1)
        got_backup2 = resolver._keep_backup_of_snapshot(snapshot2)

    assert got_backup1 == backup1
    assert got_backup2 == backup2

    assert resolver.get_result().keep_backups == {
        backup1.uuid: KeepBackup(backup1, KeepMeta(reasons=Reasons.Preserved)),
        backup2.uuid: KeepBackup(backup2, KeepMeta(reasons=Reasons.Preserved)),
    }



from __future__ import annotations

from random import randrange
from typing import TYPE_CHECKING
from uuid import uuid4

import arrow
from btrfs2s3._internal.backups import BackupInfo
from btrfs2s3._internal.s3 import iter_backups

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client


def test_empty(s3: S3Client, bucket: str) -> None:
    got = list(iter_backups(s3, bucket))
    assert got == []


def test_objects_but_none_are_backups(s3: S3Client, bucket: str) -> None:
    s3.put_object(Bucket=bucket, Key="not-a-bucket-name.gz", Body=b"dummy")
    got = list(iter_backups(s3, bucket))
    assert got == []


def mkinfo() -> BackupInfo:
    return BackupInfo(
        uuid=uuid4().bytes,
        parent_uuid=uuid4().bytes,
        ctransid=randrange(100000),
        ctime=arrow.get().timestamp(),
        send_parent_uuid=uuid4().bytes,
    )


def test_one_backup(s3: S3Client, bucket: str) -> None:
    info = mkinfo()
    key = f"basename{''.join(info.get_path_suffixes())}"
    s3.put_object(Bucket=bucket, Key=key, Body=b"dummy")

    got = list(iter_backups(s3, bucket))

    assert len(got) == 1
    got_object, got_info = got[0]
    assert got_info == info
    assert got_object["Key"] == key


def test_pagination(s3: S3Client, bucket: str) -> None:
    infos = {mkinfo() for _ in range(1001)}
    for info in infos:
        key = f"basename{''.join(info.get_path_suffixes())}"
        s3.put_object(Bucket=bucket, Key=key, Body=b"dummy")

    got = list(iter_backups(s3, bucket))

    got_infos = {info for _, info in got}
    assert got_infos == infos



from __future__ import annotations

import time
from typing import TYPE_CHECKING
from uuid import uuid4

from botocore.exceptions import ClientError
from btrfs2s3._internal.action import Actions
from btrfs2s3._internal.assessor import assess
from btrfs2s3._internal.assessor import assessment_to_actions
from btrfs2s3._internal.backups import BackupInfo
from btrfs2s3._internal.preservation import Params
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.s3 import iter_backups
from btrfs2s3._internal.util import backup_of_snapshot
from btrfs2s3._internal.util import NULL_UUID
from btrfs2s3._internal.util import SubvolumeFlags
import btrfsutil
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from mypy_boto3_s3.client import S3Client

    from tests.conftest import DownloadAndPipe


def test_create_and_backup_new_snapshot(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    btrfsutil.sync(source)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )

    (source_asmt,) = list(assessment.sources.values())
    # We should only keep one proposed snapshot
    (snapshot_asmt,) = list(source_asmt.snapshots.values())
    proposed_info = snapshot_asmt.info
    assert proposed_info.uuid != NULL_UUID
    assert proposed_info.parent_uuid == btrfsutil.subvolume_info(source).uuid
    want_flags = SubvolumeFlags.Proposed | SubvolumeFlags.ReadOnly
    assert proposed_info.flags & want_flags == want_flags
    assert snapshot_asmt.keep_meta == KeepMeta(
        reasons=Reasons.MostRecent, flags=Flags.New
    )
    assert snapshot_asmt.new

    # We should only keep one proposed full backup
    ((backup_uuid, backup_asmt),) = list(source_asmt.backups.items())
    assert backup_uuid == proposed_info.uuid
    assert backup_asmt.keep_meta == KeepMeta(
        reasons=Reasons.MostRecent, flags=Flags.New
    )
    assert backup_asmt.new

    actions = Actions()
    assessment_to_actions(assessment, actions)

    (create_snapshot_intent,) = list(actions.iter_create_snapshot_intents())
    assert create_snapshot_intent.source.peek() == source
    assert snapshot_dir in create_snapshot_intent.path().parents
    assert create_snapshot_intent.path().name.startswith(source.name)
    (rename_snapshot_intent,) = list(actions.iter_rename_snapshot_intents())
    assert rename_snapshot_intent.source.peek() == create_snapshot_intent.path.peek()
    (create_backup_intent,) = list(actions.iter_create_backup_intents())
    assert create_backup_intent.source.peek() == source

    actions.execute(s3, bucket)

    (snapshot,) = list(snapshot_dir.iterdir())
    assert snapshot.name.startswith(source.name)
    snapshot_info = btrfsutil.subvolume_info(snapshot)
    assert snapshot_info.parent_uuid == btrfsutil.subvolume_info(source).uuid

    ((obj, backup),) = list(iter_backups(s3, bucket))
    assert backup.uuid == snapshot_info.uuid
    assert backup.parent_uuid == snapshot_info.parent_uuid
    assert backup.send_parent_uuid is None

    download_and_pipe(obj["Key"], ["btrfs", "receive", "--dump"])


def test_create_and_backup_with_parent(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    # Create an initial snapshot
    snapshot1 = snapshot_dir / "snapshot1"
    btrfsutil.create_snapshot(source, snapshot1, read_only=True)
    # Modify the source again
    (source / "dummy-file").write_bytes(b"dummy2")
    btrfsutil.sync(source)

    # This isn't guaranteed to work at year boundaries. Can't think of a better
    # way to do it right now.
    now = time.time()
    policy = Policy(now=now, params=Params(years=1))
    expected_time_span = next(policy.iter_time_spans(now))
    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=policy,
    )

    (source_asmt,) = list(assessment.sources.values())
    (old_asmt, new_asmt) = sorted(source_asmt.snapshots.values(), key=lambda a: a.new)
    # One assessment for the new snapshot
    proposed_info = new_asmt.info
    assert proposed_info.uuid != NULL_UUID
    assert proposed_info.parent_uuid == btrfsutil.subvolume_info(source).uuid
    want_flags = SubvolumeFlags.Proposed | SubvolumeFlags.ReadOnly
    assert proposed_info.flags & want_flags == want_flags
    assert new_asmt.keep_meta == KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New)
    assert new_asmt.new
    # One assessment for the existing snapshot
    assert old_asmt.info.uuid == btrfsutil.subvolume_info(snapshot1).uuid
    assert old_asmt.keep_meta == KeepMeta(
        reasons=Reasons.Preserved, time_spans={expected_time_span}
    )
    assert not old_asmt.new

    # One backup assessment will match the uuid of the existing snapshot, but
    # the other will match a proposed uuid
    full_asmt = source_asmt.backups[old_asmt.info.uuid]
    delta_asmt = source_asmt.backups[new_asmt.info.uuid]
    assert full_asmt.backup.check().uuid == old_asmt.info.uuid
    assert full_asmt.backup.check().send_parent_uuid is None
    assert full_asmt.keep_meta == KeepMeta(
        reasons=Reasons.Preserved, flags=Flags.New, time_spans={expected_time_span}
    )
    assert full_asmt.new
    assert delta_asmt.keep_meta == KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New)
    assert delta_asmt.new

    actions = Actions()
    assessment_to_actions(assessment, actions)

    actions.execute(s3, bucket)

    (snapshot1, snapshot2) = sorted(snapshot_dir.iterdir())
    assert snapshot1.name.startswith(source.name)
    assert snapshot2.name.startswith(source.name)
    snapshot1_info = btrfsutil.subvolume_info(snapshot1)
    snapshot2_info = btrfsutil.subvolume_info(snapshot2)
    assert snapshot1_info.parent_uuid == btrfsutil.subvolume_info(source).uuid
    assert snapshot2_info.parent_uuid == btrfsutil.subvolume_info(source).uuid

    ((full_obj, full_backup), (delta_obj, delta_backup)) = sorted(
        iter_backups(s3, bucket), key=lambda x: x[1].ctransid
    )
    assert full_backup.uuid == snapshot1_info.uuid
    assert full_backup.send_parent_uuid is None
    assert delta_backup.uuid == snapshot2_info.uuid
    assert delta_backup.send_parent_uuid == full_backup.uuid

    download_and_pipe(full_obj["Key"], ["btrfs", "receive", "--dump"])
    download_and_pipe(delta_obj["Key"], ["btrfs", "receive", "--dump"])


def test_rename_snapshot(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    # Snapshot dir
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # One snapshot, but wrongly-named
    snapshot = snapshot_dir / "initial-snapshot-name-that-needs-to-be-updated"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)
    btrfsutil.sync(source)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )

    (source_asmt,) = list(assessment.sources.values())
    # Keep only the existing snapshot
    (snapshot_asmt,) = list(source_asmt.snapshots.values())
    assert snapshot_asmt.info.uuid == btrfsutil.subvolume_info(snapshot).uuid
    assert snapshot_asmt.info.parent_uuid == btrfsutil.subvolume_info(source).uuid
    check_flags = SubvolumeFlags.Proposed | SubvolumeFlags.ReadOnly
    want_flags = SubvolumeFlags.ReadOnly
    assert snapshot_asmt.info.flags & check_flags == want_flags
    assert snapshot_asmt.target_path.check().name.startswith(source.name)
    assert snapshot_asmt.keep_meta == KeepMeta(reasons=Reasons.MostRecent)
    assert not snapshot_asmt.new

    # We should only keep one full backup, of the existing snapshot
    ((backup_uuid, backup_asmt),) = list(source_asmt.backups.items())
    assert backup_uuid == snapshot_asmt.info.uuid
    assert backup_asmt.keep_meta == KeepMeta(
        reasons=Reasons.MostRecent, flags=Flags.New
    )
    assert backup_asmt.new

    actions = Actions()
    assessment_to_actions(assessment, actions)

    (rename_snapshot_intent,) = list(actions.iter_rename_snapshot_intents())
    assert rename_snapshot_intent.source.peek() == snapshot
    (create_backup_intent,) = list(actions.iter_create_backup_intents())
    assert create_backup_intent.source.peek() == source

    actions.execute(s3, bucket)

    (snapshot,) = list(snapshot_dir.iterdir())
    assert snapshot.name.startswith(source.name)
    snapshot_info = btrfsutil.subvolume_info(snapshot)
    assert snapshot_info.parent_uuid == btrfsutil.subvolume_info(source).uuid

    ((obj, backup),) = list(iter_backups(s3, bucket))
    assert backup.uuid == snapshot_info.uuid
    assert backup.parent_uuid == snapshot_info.parent_uuid
    assert backup.send_parent_uuid is None

    download_and_pipe(obj["Key"], ["btrfs", "receive", "--dump"])


def test_delete_only_snapshot_because_proposed_would_be_newer(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    # Initial snapshot, name doesn't matter
    initial_snapshot = snapshot_dir / "initial-snapshot"
    btrfsutil.create_snapshot(source, initial_snapshot, read_only=True)
    # Modify the source again
    (source / "dummy-file").write_bytes(b"dummy2")
    btrfsutil.sync(source)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )

    (source_asmt,) = list(assessment.sources.values())
    # Two assessments: one for the existing snapshot, one proposed
    (old_asmt, new_asmt) = sorted(source_asmt.snapshots.values(), key=lambda a: a.new)
    proposed_info = new_asmt.info
    assert proposed_info.uuid != NULL_UUID
    assert proposed_info.parent_uuid == btrfsutil.subvolume_info(source).uuid
    want_flags = SubvolumeFlags.Proposed | SubvolumeFlags.ReadOnly
    assert proposed_info.flags & want_flags == want_flags
    assert new_asmt.keep_meta == KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New)
    assert new_asmt.new
    # Assessment of the existing snapshot
    assert old_asmt.info.uuid == btrfsutil.subvolume_info(initial_snapshot).uuid
    assert old_asmt.initial_path == initial_snapshot
    assert old_asmt.keep_meta == KeepMeta()
    assert not old_asmt.new

    # We should only keep one proposed full backup
    ((backup_uuid, backup_asmt),) = list(source_asmt.backups.items())
    assert backup_uuid == proposed_info.uuid
    assert backup_asmt.keep_meta == KeepMeta(
        reasons=Reasons.MostRecent, flags=Flags.New
    )
    assert backup_asmt.new

    actions = Actions()
    assessment_to_actions(assessment, actions)

    (create_snapshot_intent,) = list(actions.iter_create_snapshot_intents())
    assert create_snapshot_intent.source.peek() == source
    assert snapshot_dir in create_snapshot_intent.path().parents
    assert create_snapshot_intent.path().name.startswith(source.name)
    (rename_snapshot_intent,) = list(actions.iter_rename_snapshot_intents())
    assert rename_snapshot_intent.source.peek() == create_snapshot_intent.path.peek()
    (create_backup_intent,) = list(actions.iter_create_backup_intents())
    assert create_backup_intent.source.peek() == source
    (delete_snapshot_intent,) = list(actions.iter_delete_snapshot_intents())
    assert delete_snapshot_intent.path.peek() == initial_snapshot

    actions.execute(s3, bucket)

    (snapshot,) = list(snapshot_dir.iterdir())
    assert snapshot.name.startswith(source.name)
    snapshot_info = btrfsutil.subvolume_info(snapshot)
    assert snapshot_info.parent_uuid == btrfsutil.subvolume_info(source).uuid

    ((obj, backup),) = list(iter_backups(s3, bucket))
    assert backup.uuid == snapshot_info.uuid
    assert backup.parent_uuid == snapshot_info.parent_uuid
    assert backup.send_parent_uuid is None

    download_and_pipe(obj["Key"], ["btrfs", "receive", "--dump"])


def test_ignore_read_write_snapshot(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    # Initial snapshot, name doesn't matter
    initial_snapshot = snapshot_dir / "initial-snapshot"
    btrfsutil.create_snapshot(source, initial_snapshot)
    btrfsutil.sync(source)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    actions.execute(s3, bucket)

    (snap1, snap2) = sorted(snapshot_dir.iterdir(), key=lambda p: p == initial_snapshot)
    assert snap2 == initial_snapshot
    info = btrfsutil.subvolume_info(snap1)
    assert info.flags & SubvolumeFlags.ReadOnly


def test_fail_if_snapshot_dir_not_on_btrfs(
    btrfs_mountpoint: Path, ext4_mountpoint: Path, s3: S3Client, bucket: str
) -> None:
    with pytest.raises(RuntimeError):
        assess(
            snapshot_dir=ext4_mountpoint,
            sources=(btrfs_mountpoint,),
            s3=s3,
            bucket=bucket,
            policy=Policy(),
        )


def test_ignore_snapshots_from_unrelated_sources(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str
) -> None:
    # Create subvolumes
    source1 = btrfs_mountpoint / "source1"
    btrfsutil.create_subvolume(source1)
    source2 = btrfs_mountpoint / "source2"
    btrfsutil.create_subvolume(source2)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the sources
    (source1 / "dummy-file").write_bytes(b"dummy")
    (source2 / "dummy-file").write_bytes(b"dummy")
    # Initial snapshot, name doesn't matter
    ignore_snapshot = snapshot_dir / "ignore-snapshot"
    btrfsutil.create_snapshot(source2, ignore_snapshot, read_only=True)
    btrfsutil.sync(btrfs_mountpoint)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source1,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    actions.execute(s3, bucket)

    (snap1, snap2) = sorted(snapshot_dir.iterdir(), key=lambda p: p == ignore_snapshot)
    assert snap2 == ignore_snapshot
    info = btrfsutil.subvolume_info(snap1)
    assert info.flags & SubvolumeFlags.ReadOnly


def test_ignore_unrelated_s3_objects(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str
) -> None:
    key = "not-a-backup-name"
    s3.put_object(Bucket=bucket, Key=key, Body=b"dummy")

    assessment = assess(
        snapshot_dir=btrfs_mountpoint,
        sources=(btrfs_mountpoint,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    actions.execute(s3, bucket)

    assert s3.get_object(Bucket=bucket, Key=key)["Body"].read() == b"dummy"


def test_ignore_unrelated_backups(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str
) -> None:
    info = BackupInfo(
        uuid=uuid4().bytes,
        parent_uuid=uuid4().bytes,
        send_parent_uuid=None,
        ctransid=1234,
        ctime=time.time(),
    )
    key = f"base{''.join(info.get_path_suffixes())}"
    s3.put_object(Bucket=bucket, Key=key, Body=b"dummy")

    assessment = assess(
        snapshot_dir=btrfs_mountpoint,
        sources=(btrfs_mountpoint,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    actions.execute(s3, bucket)

    assert s3.get_object(Bucket=bucket, Key=key)["Body"].read() == b"dummy"


def test_delete_old_backups(btrfs_mountpoint: Path, s3: S3Client, bucket: str) -> None:
    # Create subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Initial snapshot, name doesn't matter
    dummy_snapshot = snapshot_dir / "dummy-snapshot"
    btrfsutil.create_snapshot(source, dummy_snapshot, read_only=True)
    # Create a dummy backup to match the snapshot
    dummy_backup = backup_of_snapshot(btrfsutil.subvolume_info(dummy_snapshot))
    dummy_key = f"base{''.join(dummy_backup.get_path_suffixes())}"
    s3.put_object(Bucket=bucket, Key=dummy_key, Body=b"dummy")
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    btrfsutil.sync(btrfs_mountpoint)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    actions.execute(s3, bucket)

    # Ensure dummy backup was deleted
    with pytest.raises(ClientError):
        s3.head_object(Bucket=bucket, Key=dummy_key)


def test_second_run_is_a_no_op(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str
) -> None:
    # Create subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    btrfsutil.sync(btrfs_mountpoint)

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    actions.execute(s3, bucket)

    # Second run

    assessment = assess(
        snapshot_dir=snapshot_dir,
        sources=(source,),
        s3=s3,
        bucket=bucket,
        policy=Policy(),
    )
    actions = Actions()
    assessment_to_actions(assessment, actions)
    assert actions.empty()



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

from btrfs2s3._internal.console import THEME
from btrfs2s3._internal.main import main
from btrfs2s3._internal.s3 import iter_backups
import btrfsutil
from rich.console import Console

if TYPE_CHECKING:
    from pathlib import Path

    from mypy_boto3_s3.client import S3Client
    import pytest

    from tests.conftest import DownloadAndPipe


def test_pretend(
    tmp_path: Path,
    btrfs_mountpoint: Path,
    bucket: str,
    capsys: pytest.CaptureFixture[str],
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    btrfsutil.sync(source)

    console = Console(force_terminal=True, theme=THEME, width=88, height=30)
    config_path = tmp_path / "config.yaml"
    config_path.write_text(f"""
      timezone: UTC
      sources:
      - path: {source}
        snapshots: {snapshot_dir}
        upload_to_remotes:
        - id: aws
          preserve: 1y
      remotes:
      - id: aws
        s3:
          bucket: {bucket}
    """)
    assert main(console=console, argv=["update", "--pretend", str(config_path)]) == 0

    (out, err) = capsys.readouterr()
    # No idea how to stabilize this for golden testing
    assert "assessment and proposed new state" in out
    assert err == ""


def test_force(
    tmp_path: Path,
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    capsys: pytest.CaptureFixture[str],
    download_and_pipe: DownloadAndPipe,
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    btrfsutil.sync(source)

    console = Console(force_terminal=True, theme=THEME, width=88, height=30)
    config_path = tmp_path / "config.yaml"
    config_path.write_text(f"""
      timezone: UTC
      sources:
      - path: {source}
        snapshots: {snapshot_dir}
        upload_to_remotes:
        - id: aws
          preserve: 1y
      remotes:
      - id: aws
        s3:
          bucket: {bucket}
    """)
    argv = ["update", "--force", str(config_path)]
    assert main(console=console, argv=argv) == 0

    (out, err) = capsys.readouterr()
    # No idea how to stabilize this for golden testing
    assert "assessment and proposed new state" in out
    assert err == ""

    (snapshot,) = snapshot_dir.iterdir()
    info = btrfsutil.subvolume_info(snapshot)
    assert info.parent_uuid == btrfsutil.subvolume_info(source).uuid
    ((obj, backup),) = iter_backups(s3, bucket)
    assert backup.uuid == info.uuid
    assert backup.send_parent_uuid is None
    download_and_pipe(obj["Key"], ["btrfs", "receive", "--dump"])

    # Second run should be no-op
    assert main(argv=argv) == 0

    (out, err) = capsys.readouterr()
    assert "nothing to be done" in out


def test_refuse_to_run_unattended_without_pretend_or_force(
    tmp_path: Path, goldifyconsole: Console
) -> None:
    # This shouldn't get to the point of verifying arguments
    config_path = tmp_path / "config.yaml"
    config_path.write_text("""
      timezone: UTC
      sources:
      - path: dummy_source
        snapshots: dummy_snapshot_dir
        upload_to_remotes:
        - id: aws
          preserve: 1y
      remotes:
      - id: aws
        s3:
          bucket: dummy_bucket
    """)
    assert main(argv=["update", str(config_path)], console=goldifyconsole) == 1


def test_reject_continue_prompt(
    tmp_path: Path,
    btrfs_mountpoint: Path,
    bucket: str,
    capsys: pytest.CaptureFixture[str],
    s3: S3Client,
) -> None:
    # Create a subvolume
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    # Snapshot dir, but no snapshots
    snapshot_dir = btrfs_mountpoint / "snapshots"
    snapshot_dir.mkdir()
    # Modify some data in the source
    (source / "dummy-file").write_bytes(b"dummy")
    btrfsutil.sync(source)

    console = Console(force_terminal=True, theme=THEME, width=88, height=30)
    config_path = tmp_path / "config.yaml"
    config_path.write_text(f"""
      timezone: UTC
      sources:
      - path: {source}
        snapshots: {snapshot_dir}
        upload_to_remotes:
        - id: aws
          preserve: 1y
      remotes:
      - id: aws
        s3:
          bucket: {bucket}
    """)
    with patch("rich.console.input", return_value="n"):
        assert main(console=console, argv=["update", str(config_path)]) == 0

    (out, err) = capsys.readouterr()
    # No idea how to stabilize this for golden testing
    assert "continue?" in out
    assert err == ""

    # Ensure there were no side effects
    assert list(snapshot_dir.iterdir()) == []
    assert s3.list_objects_v2(Bucket=bucket).get("Contents", []) == []



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

import arrow
from botocore.exceptions import ClientError
from btrfs2s3._internal.action import Actions
from btrfs2s3._internal.util import backup_of_snapshot
import btrfsutil
from btrfsutil import SubvolumeInfo
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from mypy_boto3_s3.client import S3Client

    from tests.conftest import DownloadAndPipe


def test_empty() -> None:
    actions = Actions()
    assert actions.empty()


def test_create_and_rename(btrfs_mountpoint: Path, s3: S3Client, bucket: str) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    (source / "dummy-file").write_bytes(b"dummy")

    actions = Actions()

    initial_path = btrfs_mountpoint / "snapshot"

    @functools.lru_cache
    def get_info() -> SubvolumeInfo:
        return btrfsutil.subvolume_info(initial_path)

    def get_target_path() -> Path:
        ctime = arrow.get(get_info().ctime)
        return btrfs_mountpoint / f"snapshot-{ctime.isoformat()}"

    actions.create_snapshot(source=source, path=initial_path)
    actions.rename_snapshot(source=initial_path, target=get_target_path)
    assert not actions.empty()

    actions.execute(s3, bucket)

    # Ensure the snapshot was created and renamed
    assert (get_target_path() / "dummy-file").read_bytes() == b"dummy"


def test_create_rename_backup(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    (source / "dummy-file").write_bytes(b"dummy")
    snapshot1 = btrfs_mountpoint / "snapshot1"
    btrfsutil.create_snapshot(source, snapshot1, read_only=True)
    (source / "dummy-file").write_bytes(b"dummy2")

    actions = Actions()

    initial_path = btrfs_mountpoint / "snapshot2"

    @functools.lru_cache
    def get_info() -> SubvolumeInfo:
        return btrfsutil.subvolume_info(initial_path)

    def get_target_path() -> Path:
        ctime = arrow.get(get_info().ctime)
        return btrfs_mountpoint / f"snapshot-{ctime.isoformat()}"

    def get_send_parent() -> Path | None:
        return snapshot1

    def get_key() -> str:
        backup = backup_of_snapshot(
            get_info(), send_parent=btrfsutil.subvolume_info(snapshot1)
        )
        return f"{source.name}{''.join(backup.get_path_suffixes())}"

    actions.create_snapshot(source=source, path=initial_path)
    actions.rename_snapshot(source=initial_path, target=get_target_path)
    actions.create_backup(
        source=source,
        snapshot=get_target_path,
        send_parent=get_send_parent,
        key=get_key,
    )
    assert not actions.empty()

    actions.execute(s3, bucket)

    # Ensure the snapshot was created and renamed
    assert (get_target_path() / "dummy-file").read_bytes() == b"dummy2"

    # Test the backup exists and is a valid btrfs archive
    download_and_pipe(get_key(), ["btrfs", "receive", "--dump"])


def test_other_actions(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    def mksnapshot(name: str) -> Path:
        path = btrfs_mountpoint / name
        btrfsutil.create_snapshot(source, path, read_only=True)
        return path

    actions = Actions()

    # Rename a snapshot
    rename_me = mksnapshot("rename-me")
    rename_to = rename_me.parent / "rename-to"
    actions.rename_snapshot(source=rename_me, target=rename_to)

    # Delete a snapshot
    delete_me = mksnapshot("delete-me")
    actions.delete_snapshot(delete_me)

    # Create a backup
    backup_of_me = mksnapshot("backup-of-me")
    backup_of_me_key = "backup-key"
    actions.create_backup(
        source=source, snapshot=backup_of_me, send_parent=None, key=backup_of_me_key
    )

    # Delete a backup
    delete_me_key = "delete-me-key"
    s3.put_object(Bucket=bucket, Key=delete_me_key, Body=b"dummy")
    actions.delete_backup(delete_me_key)
    assert not actions.empty()

    actions.execute(s3, bucket)

    # Check rename
    assert not rename_me.exists()
    assert rename_to.exists()

    # Check delete
    assert not delete_me.exists()

    # Check backup
    download_and_pipe(backup_of_me_key, ["btrfs", "receive", "--dump"])

    # Check delete
    with pytest.raises(ClientError):
        s3.head_object(Bucket=bucket, Key=delete_me_key)



from __future__ import annotations

import collections
import contextlib
from contextvars import ContextVar
import dataclasses
from dataclasses import field
import enum
from queue import SimpleQueue
from typing import Generic
from typing import Protocol
from typing import TYPE_CHECKING
import uuid
import warnings

from btrfsutil import SubvolumeInfo
from typing_extensions import Self
from typing_extensions import TypeAlias
from typing_extensions import TypeVar

from btrfs2s3._internal.backups import BackupInfo
from btrfs2s3._internal.util import backup_of_snapshot

if TYPE_CHECKING:
    from typing import Collection
    from typing import Iterator

    from btrfs2s3._internal.preservation import Policy
    from btrfs2s3._internal.preservation import TS


class _Info(Protocol):
    @property
    def uuid(self) -> bytes: ...
    @property
    def ctime(self) -> float: ...
    @property
    def ctransid(self) -> int: ...


_I = TypeVar("_I", bound=_Info)


class _Index(Generic[_I]):
    def __init__(self, *, items: Collection[_I], policy: Policy) -> None:
        self._item_by_uuid = {}
        self._items_by_time_span = collections.defaultdict(list)

        for item in items:
            self._item_by_uuid[item.uuid] = item
            for time_span in policy.iter_time_spans(item.ctime):
                self._items_by_time_span[time_span].append(item)

    def get_nominal(self, time_span: TS) -> _I | None:
        items = self._items_by_time_span.get(time_span)
        if items:
            return min(items, key=lambda i: i.ctransid)
        return None

    def get(self, uuid: bytes) -> _I | None:
        return self._item_by_uuid.get(uuid)

    def get_most_recent(self) -> _I | None:
        if self._item_by_uuid:
            return max(self._item_by_uuid.values(), key=lambda i: i.ctransid)
        return None

    def get_all_time_spans(self) -> Collection[TS]:
        return self._items_by_time_span.keys()


class Reasons(enum.Flag):
    Empty = 0
    Preserved = enum.auto()
    MostRecent = enum.auto()
    SendAncestor = enum.auto()


class Flags(enum.Flag):
    Empty = 0
    New = enum.auto()
    ReplacingNewer = enum.auto()
    NoSnapshot = enum.auto()
    SnapshotIsNewer = enum.auto()


@dataclasses.dataclass
class KeepMeta:
    reasons: Reasons = Reasons.Empty
    flags: Flags = Flags.Empty
    time_spans: set[TS] = field(default_factory=set)
    other_uuids: set[bytes] = field(default_factory=set)

    def __or__(self, other: Self) -> Self:
        return self.__class__(
            reasons=self.reasons | other.reasons,
            flags=self.flags | other.flags,
            time_spans=self.time_spans | other.time_spans,
            other_uuids=self.other_uuids | other.other_uuids,
        )


@dataclasses.dataclass
class _MarkedItem(Generic[_I]):
    item: _I
    meta: KeepMeta = field(default_factory=KeepMeta)


class _Marker(Generic[_I]):
    def __init__(self) -> None:
        self._result: dict[bytes, _MarkedItem[_I]] = {}
        self._reasons_ctx: ContextVar[Reasons] = ContextVar(
            "reasons", default=Reasons.Empty
        )
        self._flags_ctx: ContextVar[Flags] = ContextVar("flags", default=Flags.Empty)
        self._time_span_ctx: ContextVar[set[TS]] = ContextVar(
            "time_span", default=set()
        )

    @contextlib.contextmanager
    def with_reasons(self, reasons: Reasons) -> Iterator[None]:
        existing = self._reasons_ctx.get()
        token = self._reasons_ctx.set(reasons | existing)
        try:
            yield
        finally:
            self._reasons_ctx.reset(token)

    @contextlib.contextmanager
    def with_time_span(self, time_span: TS) -> Iterator[None]:
        existing = self._time_span_ctx.get()
        token = self._time_span_ctx.set(existing | {time_span})
        try:
            yield
        finally:
            self._time_span_ctx.reset(token)

    def get_result(self) -> dict[bytes, _MarkedItem[_I]]:
        return self._result

    def mark(
        self, item: _I, *, flags: Flags = Flags.Empty, other_uuid: bytes | None = None
    ) -> None:
        if item.uuid not in self._result:
            self._result[item.uuid] = _MarkedItem(item=item)
        marked = self._result[item.uuid]
        update = KeepMeta(
            reasons=self._reasons_ctx.get(),
            flags=flags | self._flags_ctx.get(),
            time_spans=self._time_span_ctx.get(),
            other_uuids={other_uuid} if other_uuid is not None else set(),
        )
        if not update.reasons:
            raise AssertionError
        marked.meta |= update


KeepSnapshot: TypeAlias = _MarkedItem[SubvolumeInfo]
KeepBackup: TypeAlias = _MarkedItem[BackupInfo]


@dataclasses.dataclass(frozen=True)
class Result:
    keep_snapshots: dict[bytes, KeepSnapshot] = field(default_factory=dict)
    keep_backups: dict[bytes, KeepBackup] = field(default_factory=dict)


class _Resolver:
    def __init__(
        self,
        *,
        snapshots: Collection[SubvolumeInfo],
        backups: Collection[BackupInfo],
        policy: Policy,
    ) -> None:
        self._snapshots = _Index(items=snapshots, policy=policy)
        self._backups = _Index(items=backups, policy=policy)
        self._policy = policy

        self._keep_snapshots: _Marker[SubvolumeInfo] = _Marker()
        self._keep_backups: _Marker[BackupInfo] = _Marker()

    @contextlib.contextmanager
    def _with_time_span(self, time_span: TS) -> Iterator[None]:
        with self._keep_snapshots.with_time_span(time_span):  # noqa: SIM117
            with self._keep_backups.with_time_span(time_span):
                yield

    @contextlib.contextmanager
    def _with_reasons(self, reasons: Reasons) -> Iterator[None]:
        with self._keep_snapshots.with_reasons(reasons):  # noqa: SIM117
            with self._keep_backups.with_reasons(reasons):
                yield

    def get_result(self) -> Result:
        return Result(
            keep_snapshots=self._keep_snapshots.get_result(),
            keep_backups=self._keep_backups.get_result(),
        )

    def _keep_backup_of_snapshot(
        self,
        snapshot: SubvolumeInfo,
        *,
        flags: Flags = Flags.Empty,
        other_uuid: bytes | None = None,
    ) -> BackupInfo:
        backup: BackupInfo | None = None
        # Use an existing backup when available
        if snapshot.uuid in self._keep_backups.get_result():
            backup = self._keep_backups.get_result()[snapshot.uuid].item
        if backup is None:
            backup = self._backups.get(snapshot.uuid)
        if backup is None:
            flags |= Flags.New
            # Determine send-parent for a new backup
            send_parent: SubvolumeInfo | None = None
            for snapshot_time_span in self._policy.iter_time_spans(snapshot.ctime):
                nominal_snapshot = self._snapshots.get_nominal(snapshot_time_span)
                if nominal_snapshot is None:
                    raise AssertionError
                if nominal_snapshot.uuid == snapshot.uuid:
                    break
                send_parent = nominal_snapshot

            backup = backup_of_snapshot(snapshot, send_parent=send_parent)

        self._keep_backups.mark(backup, flags=flags, other_uuid=other_uuid)
        return backup

    def _keep_snapshot_and_backup_for_time_span(self, time_span: TS) -> None:
        nominal_snapshot = self._snapshots.get_nominal(time_span)
        nominal_backup = self._backups.get_nominal(time_span)

        # NB: we could have a nominal backup older than our nominal snapshot.
        # If so, we keep the existing backup, but don't back up the snapshot.
        # There's some duplication in the following logic, but the duplicated
        # cases are expected to be rare and this makes it easier to follow
        if nominal_snapshot:
            self._keep_snapshots.mark(nominal_snapshot)
            if nominal_backup is None:
                self._keep_backup_of_snapshot(nominal_snapshot, flags=Flags.New)
            elif nominal_backup.ctransid > nominal_snapshot.ctransid:
                self._keep_backup_of_snapshot(
                    nominal_snapshot, flags=Flags.ReplacingNewer
                )
        if nominal_backup:
            if nominal_snapshot is None:
                self._keep_backups.mark(nominal_backup, flags=Flags.NoSnapshot)
            elif nominal_backup.ctransid < nominal_snapshot.ctransid:
                self._keep_backups.mark(nominal_backup, flags=Flags.SnapshotIsNewer)
            elif nominal_backup.ctransid == nominal_snapshot.ctransid:
                self._keep_backups.mark(nominal_backup)

    def keep_snapshots_and_backups_for_preserved_time_spans(self) -> None:
        with self._with_reasons(Reasons.Preserved):
            all_time_spans = set(self._snapshots.get_all_time_spans()) | set(
                self._backups.get_all_time_spans()
            )
            for time_span in all_time_spans:
                if not self._policy.should_preserve_for_time_span(time_span):
                    continue
                with self._with_time_span(time_span):
                    self._keep_snapshot_and_backup_for_time_span(time_span)

    def _keep_most_recent_snapshot(self) -> None:
        most_recent_snapshot = self._snapshots.get_most_recent()
        if most_recent_snapshot:
            self._keep_snapshots.mark(most_recent_snapshot)
            self._keep_backup_of_snapshot(most_recent_snapshot)

    def keep_most_recent_snapshot(self) -> None:
        with self._with_reasons(Reasons.MostRecent):
            self._keep_most_recent_snapshot()

    def _keep_send_ancestors_of_backups(self) -> None:
        # Ensure the send-parent ancestors of any kept backups are also kept
        backups_to_check: SimpleQueue[BackupInfo] = SimpleQueue()
        for marked_item in self._keep_backups.get_result().values():
            backups_to_check.put(marked_item.item)
        while not backups_to_check.empty():
            backup = backups_to_check.get()
            if not backup.send_parent_uuid:
                continue
            if backup.send_parent_uuid in self._keep_backups.get_result():
                continue
            parent_backup = self._backups.get(backup.send_parent_uuid)
            if parent_backup:
                # This can be common. For example in January, if December's monthly
                # backup should be preserved, but the December backup's send-parent
                # is last year's yearly backup
                self._keep_backups.mark(parent_backup, other_uuid=backup.uuid)
            else:
                parent_snapshot = self._snapshots.get(backup.send_parent_uuid)
                if parent_snapshot:
                    parent_backup = self._keep_backup_of_snapshot(
                        parent_snapshot, flags=Flags.New, other_uuid=backup.uuid
                    )
                else:
                    warnings.warn(
                        f"Backup chain is broken: {uuid.UUID(bytes=backup.uuid)} "
                        f"has parent {uuid.UUID(bytes=backup.send_parent_uuid)}, "
                        "which is missing",
                        stacklevel=1,
                    )
            if parent_backup:
                backups_to_check.put(parent_backup)

    def keep_send_ancestors_of_backups(self) -> None:
        with self._with_reasons(Reasons.SendAncestor):
            self._keep_send_ancestors_of_backups()


def resolve(
    *,
    snapshots: Collection[SubvolumeInfo],
    backups: Collection[BackupInfo],
    policy: Policy,
) -> Result:
    resolver = _Resolver(snapshots=snapshots, backups=backups, policy=policy)
    resolver.keep_snapshots_and_backups_for_preserved_time_spans()

    resolver.keep_most_recent_snapshot()

    # Future: is there a case where we need to keep a snapshot because it'll be
    # used as the send-parent of a future backup, but *isn't* otherwise kept?

    resolver.keep_send_ancestors_of_backups()
    return resolver.get_result()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from btrfs2s3._internal.action import Actions
from btrfs2s3._internal.commands.update import print_actions
import pytest

if TYPE_CHECKING:
    from rich.console import Console


@pytest.mark.parametrize(
    (
        "n_create_snapshots",
        "n_rename_snapshots",
        "n_create_backups",
        "n_delete_snapshots",
        "n_delete_backups",
    ),
    [(0, 0, 0, 0, 0), (3, 3, 3, 3, 3)],
)
def test_print_actions(
    goldifyconsole: Console,
    n_create_snapshots: int,
    n_rename_snapshots: int,
    n_create_backups: int,
    n_delete_snapshots: int,
    n_delete_backups: int,
) -> None:
    actions = Actions()
    for i in range(n_create_snapshots):
        actions.create_snapshot(
            source=Path(f"/path/to/source{i}"), path=Path(f"/path/to/snapshot{i}")
        )
    for i in range(n_rename_snapshots):
        actions.rename_snapshot(
            source=Path(f"/path/to/source{i}"), target=Path(f"/path/to/target{i}")
        )
    for i in range(n_create_backups):
        actions.create_backup(
            source=Path(f"/path/to/source{i}"),
            snapshot=Path(f"/path/to/snapshot{i}"),
            key=f"key{i}",
            send_parent=Path(f"/path/to/send/parent{i}"),
        )
    for i in range(n_delete_snapshots):
        actions.delete_snapshot(Path(f"/path/to/snapshot{i}"))
    for i in range(n_delete_backups):
        actions.delete_backup(f"key{i}")

    print_actions(console=goldifyconsole, actions=actions)



from __future__ import annotations

from typing import TYPE_CHECKING

from btrfs2s3._internal.config import Config
from btrfs2s3._internal.config import InvalidConfigError
from btrfs2s3._internal.config import load_from_path
from btrfs2s3._internal.config import RemoteConfig
from btrfs2s3._internal.config import S3EndpointConfig
from btrfs2s3._internal.config import S3RemoteConfig
from btrfs2s3._internal.config import SourceConfig
from btrfs2s3._internal.config import UploadToRemoteConfig
import pytest

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture()
def path(tmp_path: Path) -> Path:
    return tmp_path / "config.yaml"


def test_malformed(path: Path) -> None:
    path.write_text("malformed, bad text")
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_basic(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    config = load_from_path(path)
    assert config == Config(
        {
            "timezone": "a",
            "sources": [
                SourceConfig(
                    {
                        "path": "b",
                        "snapshots": "c",
                        "upload_to_remotes": [
                            UploadToRemoteConfig({"id": "aws", "preserve": "1y 1m"})
                        ],
                    }
                )
            ],
            "remotes": [
                RemoteConfig({"id": "aws", "s3": S3RemoteConfig({"bucket": "d"})})
            ],
        }
    )


def test_multiple_sources_with_anchors_and_refs(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: &snapshots c
          upload_to_remotes: &my_remotes
          - id: aws
            preserve: 1y 1m
        - path: otherpath
          snapshots: *snapshots
          upload_to_remotes: *my_remotes
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    config = load_from_path(path)
    source0 = config["sources"][0]
    source1 = config["sources"][1]
    assert source0["snapshots"] == source1["snapshots"]
    assert source1["upload_to_remotes"] == source1["upload_to_remotes"]


def test_s3_endpoint_config(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
        remotes:
        - id: aws
          s3:
            bucket: d
            endpoint:
              aws_access_key_id: key
              aws_secret_access_key: secret
              region_name: region
              profile_name: profile
              verify: true
              endpoint_url: https://example.com
    """)
    endpoint = load_from_path(path)["remotes"][0]["s3"]["endpoint"]
    assert endpoint == S3EndpointConfig(
        {
            "aws_access_key_id": "key",
            "aws_secret_access_key": "secret",
            "region_name": "region",
            "profile_name": "profile",
            "verify": True,
            "endpoint_url": "https://example.com",
        }
    )


def test_pipe_through(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
            pipe_through:
            - [gzip]
            - [gpg, encrypt, -r, me@example.com]
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    config = load_from_path(path)
    assert config["sources"][0]["upload_to_remotes"][0]["pipe_through"] == [
        ["gzip"],
        ["gpg", "encrypt", "-r", "me@example.com"],
    ]


def test_no_sources(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources: []
        upload_to_remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_no_remotes(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          remotes:
          - id: aws
            preserve: 1y 1m
        remotes: []
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_source_with_no_upload_to_remotes(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes: []
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_invalid_preserve(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: invalid
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_invalid_upload_to_remote_id(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: does_not_Exist
            preserve: 1y
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_no_multiple_remotes(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
        remotes:
        - id: aws
          s3:
            bucket: d
        - id: b2
          s3:
            bucket: e
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_no_multiple_snapshot_locations(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
        - path: x
          snapshots: y
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_no_multiple_preserves(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
        - path: x
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)


def test_no_multiple_pipe_throughs(path: Path) -> None:
    path.write_text("""
        timezone: a
        sources:
        - path: b
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
            pipe_through:
            - [gpg, --encrypt, -r, a@example.com]
        - path: x
          snapshots: c
          upload_to_remotes:
          - id: aws
            preserve: 1y 1m
            pipe_through:
            - [gpg, --encrypt, -r, b@example.com]
        remotes:
        - id: aws
          s3:
            bucket: d
    """)
    with pytest.raises(InvalidConfigError):
        load_from_path(path)
